from .akn import StructuredDocument


class JudgmentStructure(StructuredDocument):
    structure_type = "judgmentStructure"
    main_content_tag = "judgmentBody"


class Judgment(JudgmentStructure):
    document_type = "judgment"



"""
Cobalt's class hierarchy mimics that of the Akoma Ntoso standard. There is a single root class for
all Akoma Ntoso documents. There are subclasses for each of the Akoma Ntoso document structure types,
such as hierarchicalStructure, debateStructure, etc. Finally, there is a class for each Akoma Ntoso
document type (act, bill, judgment, etc.) that extends the corresponding structure type.
"""
from collections import OrderedDict
import re
from datetime import date

from lxml import etree, objectify
from lxml.builder import ElementMaker
from iso8601 import parse_date

from .uri import FrbrUri


ENCODING_RE = re.compile(r'encoding="[\w-]+"')

DATE_FORMAT = "%Y-%m-%d"

AKN_NAMESPACES = {
    '2.0': 'http://www.akomantoso.org/2.0',
    '3.0': 'http://docs.oasis-open.org/legaldocml/ns/akn/3.0',
}
DEFAULT_VERSION = '3.0'

# a placeholder date that indicates a null date, used in the XML where a date is required by may not be known
NULL_DATE = '0001-01-01'


def datestring(value):
    """ Format a date as an XML-suitable string. If the date is None, uses NULL_DATE.
    """
    if value is None:
        return NULL_DATE
    elif isinstance(value, str):
        return value
    else:
        return "%04d-%02d-%02d" % (value.year, value.month, value.day)


def parsedate(value):
    """ Parse an XML date string into a real date. If the value is the NULL_DATE, returns None.
    """
    if value == NULL_DATE:
        return None
    return parse_date(value).date()


# Create a new objectify parser that doesn't remove blank text nodes
objectify_parser = etree.XMLParser()
objectify_parser.set_element_class_lookup(objectify.ObjectifyElementClassLookup())


def get_maker(version=DEFAULT_VERSION):
    ns = AKN_NAMESPACES[version]
    return ElementMaker(nsmap={None: ns}, namespace=ns)


class AkomaNtosoDocument:
    """ Base class for Akoma Ntoso documents.

    :ivar root: :class:`lxml.objectify.ObjectifiedElement` root of the XML document
    :ivar namespace: primary XML namespace
    """
    _parser = objectify_parser
    # the "source" attribute used on some elements where it is required.
    # contains: name, id, url
    source = ["cobalt", "cobalt", "https://github.com/laws-africa/cobalt"]

    def __init__(self, xml=None):
        if isinstance(xml, str):
            xml = xml.encode('utf-8')
        self.parse(xml)
        self.maker = objectify.ElementMaker(annotate=False, namespace=self.namespace, nsmap=self.root.nsmap)

    def parse(self, xml, document_type=None):
        """ Parse XML and ensure it's Akoma Ntoso with a known namespace. Raises ValueError on error.
        """
        self.root = objectify.fromstring(xml, parser=self._parser)

        # ensure the root element is correct
        name = self.root.tag.split('}', 1)[-1]
        if name != 'akomaNtoso':
            raise ValueError(f"XML root element must be akomaNtoso, but got {name} instead")

        self.namespace = self.get_namespace()

    def to_xml(self, *args, encoding='utf-8', **kwargs):
        return etree.tostring(self.root, *args, encoding=encoding, **kwargs)

    def get_namespace(self):
        akn_namespaces = [ns[1] for ns in sorted(list(AKN_NAMESPACES.items()), reverse=True)]
        namespaces = list(self.root.nsmap.values())
        for ns in akn_namespaces:
            if ns in namespaces:
                return ns

        raise ValueError(f"Expected to find one of the following Akoma Ntoso XML namespaces: {', '.join(akn_namespaces)}. Only these namespaces were found: {', '.join(namespaces)}")

    def ensure_element(self, name, after, at=None, attribs=None):
        """ Helper to get an element if it exists, or create it if it doesn't.

        :param name: dotted path from `self` or `at`
        :param after: element after which to place the new element if it doesn't exist
        :param at: element at which to start looking, (defaults to self if None)
        """
        node = self.get_element(name, root=at)
        if node is None:
            # TODO: what if nodes in the path don't exist?
            node = self.make_element(name.split('.')[-1], attribs)
            after.addnext(node)

        return node

    def get_element(self, name, root=None):
        """ Lookup a dotted-path element, start at root (or self if root is None). Returns None if the element doesn't exist.
        """
        parts = name.split('.')
        # this avoids an lxml warning about testing against None
        if root is not None:
            node = root
        else:
            node = self

        for p in parts:
            try:
                node = getattr(node, p)
            except AttributeError:
                return None
        return node

    def make_element(self, elem, attribs=None):
        attribs = attribs or {}
        return getattr(self.maker, elem)(**attribs)


class StructuredDocument(AkomaNtosoDocument):
    """ Common base class for AKN documents with a known document structure.
    """

    structure_type = None
    """ The name of this document's structural type.
    """

    main_content_tag = None
    """ The name of the structural type's main content element.
    """

    document_type = None
    """ The name of the document type, corresponding to the primary document XML element.
    """

    non_eid_portions = "arguments background conclusions decision header introduction motivation preamble" \
                       " preface remedies".split()
    """ Portion names that are valid portions, but don't have eids, for use with get_portion_element.
    """

    @classmethod
    def for_document_type(cls, document_type):
        """ Return the subclass for this document type.
        """
        def check_subclasses(klass):
            for k in klass.__subclasses__():
                if k.document_type and k.document_type.lower() == document_type:
                    return k
                # recurse
                x = check_subclasses(k)
                if x:
                    return x

        document_type = document_type.lower()
        return check_subclasses(cls)

    @classmethod
    def empty_document(cls, version=DEFAULT_VERSION):
        """ Return XML for an empty document of this type, using the given AKN version.
        """
        today = datestring(date.today())
        frbr_uri = FrbrUri(
            country='za',
            locality=None,
            doctype=cls.document_type,
            subtype=None,
            date=today,
            number='1',
            work_component='main',
            language='eng',
            actor=None,
            prefix=('' if version == '2.0' else 'akn'),
        )

        maker = get_maker(version)
        content = cls.empty_document_content(maker)
        attrs = cls.empty_document_attrs()

        doc = maker.akomaNtoso(
            maker(cls.document_type,
                  cls.empty_meta(frbr_uri, maker=maker, for_root=True),
                  content,
                  **attrs)
        )
        return etree.tostring(doc, encoding='unicode')

    @classmethod
    def empty_meta(cls, frbr_uri, version=DEFAULT_VERSION, maker=None, for_root=True):
        """ Create a meta element for an frbr_uri, using the provided version or element maker.
        """
        today = datestring(date.today())
        maker = maker or get_maker(version)

        if for_root:
            # only generate top-level references if this is the root document
            refs = [maker.references(
                maker.TLCOrganization(eId=cls.source[1], href=cls.source[2], showAs=cls.source[0]),
                source=f"#{cls.source[1]}"
            )]
        else:
            refs = []

        return maker.meta(
            maker.identification(
                maker.FRBRWork(
                    maker.FRBRthis(value=frbr_uri.work_uri()),
                    maker.FRBRuri(value=frbr_uri.work_uri(work_component=False)),
                    maker.FRBRalias(value="Untitled", name="title"),
                    maker.FRBRdate(date=frbr_uri.date, name="Generation"),
                    maker.FRBRauthor(href=""),
                    maker.FRBRcountry(value=frbr_uri.place),
                    maker.FRBRnumber(value=frbr_uri.number),
                ),
                maker.FRBRExpression(
                    maker.FRBRthis(value=frbr_uri.expression_uri()),
                    maker.FRBRuri(value=frbr_uri.expression_uri(work_component=False)),
                    maker.FRBRdate(date=today, name="Generation"),
                    maker.FRBRauthor(href=""),
                    maker.FRBRlanguage(language=frbr_uri.language),
                ),
                maker.FRBRManifestation(
                    maker.FRBRthis(value=frbr_uri.manifestation_uri()),
                    maker.FRBRuri(value=frbr_uri.manifestation_uri(work_component=False)),
                    maker.FRBRdate(date=today, name="Generation"),
                    maker.FRBRauthor(href=""),
                ),
                source=f"#{cls.source[1]}"
            ),
            *refs
        )

    @classmethod
    def empty_document_content(cls, maker):
        return maker(cls.main_content_tag)

    @classmethod
    def empty_document_attrs(cls):
        return {'name': cls.document_type.lower()}

    def __init__(self, xml=None):
        """ Setup a new instance with the string in `xml`, or an empty document if the XML is not given.
        """
        if not xml:
            # use an empty document
            xml = self.empty_document()
        super().__init__(xml)

        # make, eg. ".act" an alias for ".main"
        setattr(self, self.document_type, self.main)

        # make, eg. ".body" an alias for ".main_content"
        setattr(self, self.main_content_tag, self.main_content)

    def parse(self, xml, document_type=None):
        """ Parse XML and ensure it's Akoma Ntoso.
        Raises ValueError on error. Returns the root element.
        """
        super().parse(xml, document_type)

        doc_root = self.root.find(f'./{{{self.namespace}}}{self.document_type}')
        if doc_root is None:
            raise ValueError(f"Expected {self.document_type} as a child of root element")

    @property
    def main(self):
        """ Get the main document element (normally the first child of the root element).
        """
        return getattr(self.root, self.document_type)

    @property
    def main_content(self):
        """ Get the main content element of the document.
        """
        return getattr(self.main, self.main_content_tag)

    @property
    def meta(self):
        """ Get the meta element of the document.
        """
        return self.main.meta

    @property
    def title(self):
        """ Short title """
        # look for the FRBRalias element with name="title", falling back to any alias
        title = None
        for alias in self.meta.identification.FRBRWork.iterchildren(f'{{{self.namespace}}}FRBRalias'):
            if alias.get('name') == 'title':
                return alias.get('value')
            title = alias.get('value')
        return title

    @title.setter
    def title(self, value):
        # set the title on an alias attribute with name="title"
        aliases = self.meta.identification.FRBRWork.xpath('a:FRBRalias[@name="title"]', namespaces={'a': self.namespace})
        if not aliases:
            alias = self.ensure_element('meta.identification.FRBRWork.FRBRalias', self.meta.identification.FRBRWork.FRBRuri)
            alias.set('name', 'title')
            aliases = [alias]
        aliases[0].set('value', value)

    @property
    def work_date(self):
        """ Date from the FRBRWork element. Normally, this date must match the date portion of the work's FRBR URI.
        However, since that may be a partial date (such as just a year), this return value may be different to the
        date string stored in the FRBRdate element. In particular, if the date is just a year, the month and day
        both default to 1.
        """
        return parse_date(self.meta.identification.FRBRWork.FRBRdate.get('date')).date()

    @property
    def expression_date(self):
        """ Date from the FRBRExpression element """
        return parse_date(self.meta.identification.FRBRExpression.FRBRdate.get('date')).date()

    @expression_date.setter
    def expression_date(self, value):
        self.meta.identification.FRBRExpression.FRBRdate.set('date', datestring(value))
        # set expression date of the components
        for component, element in list(self.components().items())[1:]:
            element.doc.meta.identification.FRBRExpression.FRBRdate.set('date', datestring(value))

        # update the URI
        self.frbr_uri = self.frbr_uri

    @property
    def manifestation_date(self):
        """ Date from the FRBRManifestation element """
        return parse_date(self.meta.identification.FRBRManifestation.FRBRdate.get('date')).date()

    @manifestation_date.setter
    def manifestation_date(self, value):
        self.meta.identification.FRBRManifestation.FRBRdate.set('date', datestring(value))
        # set expression date of the components
        for component, element in list(self.components().items())[1:]:
            element.doc.meta.identification.FRBRManifestation.FRBRdate.set('date', datestring(value))

    @property
    def language(self):
        """ The 3-letter ISO-639-2 language code of this document.
        """
        return self.meta.identification.FRBRExpression.FRBRlanguage.get('language', 'eng')

    @language.setter
    def language(self, value):
        self.meta.identification.FRBRExpression.FRBRlanguage.set('language', value)
        # update the URI
        self.frbr_uri = self.frbr_uri

    @property
    def frbr_uri(self):
        """ The FRBR Manifestation URI as a :class:`cobalt.uri.FrbrUri` instance that uniquely identifies this document universally.
        """
        uri = self.meta.identification.FRBRManifestation.FRBRuri.get('value')
        if uri:
            return FrbrUri.parse(uri)

    @frbr_uri.setter
    def frbr_uri(self, uri):
        if not isinstance(uri, FrbrUri):
            uri = FrbrUri.parse(uri)

        uri.language = self.meta.identification.FRBRExpression.FRBRlanguage.get('language', 'eng')
        uri.expression_date = '@' + datestring(self.expression_date)
        work_component = uri.work_component or 'main'

        # set URIs of the main document and components
        for component, element in self.components().items():
            uri.work_component = component or work_component
            ident = element.find(f'.//{{{self.namespace}}}meta/{{{self.namespace}}}identification')

            ident.FRBRWork.FRBRuri.set('value', uri.uri())
            ident.FRBRWork.FRBRthis.set('value', uri.work_uri())
            ident.FRBRWork.FRBRcountry.set('value', uri.place)
            ident.FRBRWork.FRBRdate.set('date', uri.date)

            if uri.subtype:
                self.ensure_element('FRBRsubtype', at=ident.FRBRWork, after=ident.FRBRWork.FRBRcountry).set('value', uri.subtype)
                after = ident.FRBRWork.FRBRsubtype
            else:
                after = ident.FRBRWork.FRBRcountry
                try:
                    # remove existing subtype
                    ident.FRBRWork.remove(ident.FRBRWork.FRBRsubtype)
                except AttributeError:
                    pass

            # this must come after subtype if it exists, otherwise country
            self.ensure_element('FRBRnumber', at=ident.FRBRWork, after=after).set('value', uri.number)

            ident.FRBRExpression.FRBRuri.set('value', uri.expression_uri(False))
            ident.FRBRExpression.FRBRthis.set('value', uri.expression_uri())
            ident.FRBRExpression.FRBRlanguage.set('language', uri.language)

            ident.FRBRManifestation.FRBRuri.set('value', uri.expression_uri(False))
            ident.FRBRManifestation.FRBRthis.set('value', uri.expression_uri())

    def expression_frbr_uri(self):
        """ The FRBR Expression URI as a :class:`cobalt.uri.FrbrUri` instance that uniquely identifies this document
        universally.
        """
        uri = self.meta.identification.FRBRExpression.FRBRuri.get('value')
        if uri:
            return FrbrUri.parse(uri)
        else:
            return FrbrUri.empty()

    def components(self):
        """ Get an `OrderedDict` of component name to :class:`lxml.objectify.ObjectifiedElement`
        objects. Components are this document, and `<component>` and `<attachment>` elements inside this document.
        """
        components = OrderedDict()
        frbr_uri = FrbrUri.parse(self.meta.identification.FRBRWork.FRBRthis.get('value'))
        components[frbr_uri.work_component] = self.main

        xpath = './a:attachments/a:attachment/a:*/a:meta | ./a:components/a:component/a:*/a:meta'
        for meta in self.main.xpath(xpath, namespaces={'a': self.namespace}):
            frbr_uri = FrbrUri.parse(meta.identification.FRBRWork.FRBRthis.get('value'))
            name = frbr_uri.work_component
            components[name] = meta.getparent().getparent()

        return components

    def get_portion_element(self, portion, component=None):
        """ Get a single portion of this document. The `portion` is usually an eId, as specified by
        https://docs.oasis-open.org/legaldocml/akn-nc/v1.0/os/akn-nc-v1.0-os.html#_Toc531692279.

        The optional `component` is the ancestor element within which to look for the portion.

        Range portions (eg. `chp_1->chp_3`) are not supported by this function.
        """
        root = component or self.root

        if portion in self.non_eid_portions:
            # these are valid portions that don't have eids
            xpath = f'.//a:{portion}'
        else:
            portion = portion.replace('"', '')
            xpath = f'.//a:*[@eId="{portion}"]'

        for x in root.xpath(xpath, namespaces={'a': self.namespace}):
            return x

    def _ensure_lifecycle(self):
        try:
            after = self.meta.publication
        except AttributeError:
            after = self.meta.identification
        node = self.ensure_element('meta.lifecycle', after=after)

        if not node.get('source'):
            node.set('source', '#' + self.source[1])
            self._ensure_reference('TLCOrganization', self.source[0], self.source[1], self.source[2])

        return node

    def _ensure_reference(self, elem, name, id, href):
        references = self.ensure_element('meta.references', after=self._ensure_lifecycle())

        ref = references.find(f'./{{{self.namespace}}}{elem}[@eId="{id}"]')
        if ref is None:
            ref = self.make_element(elem)
            ref.set('eId', id)
            ref.set('href', href)
            ref.set('showAs', name)
            references.insert(0, ref)
        return ref



from .akn import StructuredDocument


class CollectionStructure(StructuredDocument):
    structure_type = "collectionStructure"
    main_content_tag = "collectionBody"


class Collection(CollectionStructure):
    document_type = "documentCollection"


class OfficialGazette(CollectionStructure):
    document_type = "officialGazette"



from .akn import StructuredDocument


class PortionStructure(StructuredDocument):
    structure_type = "portionStructure"
    main_content_tag = "portionBody"


class Portion(PortionStructure):
    document_type = "portion"



"""
Cobalt can validate Akoma Ntoso documents against the Akoma Ntoso schema. Two schemas are provided:

1. Strict: the `official AKN schema <http://docs.oasis-open.org/legaldocml/akn-core/v1.0/os/part2-specs/schemas/akomantoso30.xsd>`_
2. Lenient: a slightly modified version of the official schema. Duplicate eId attributes are allowed, and FRBRdate
   elements are allowed to have year-only @date values.
"""
import os

from lxml import etree


SCHEMAS = {
    'http://docs.oasis-open.org/legaldocml/ns/akn/3.0': 'akomantoso30.xsd',
    'http://docs.oasis-open.org/legaldocml/ns/akn/3.0-lenient': 'akomantoso30-lenient.xsd',
}

_schemas = {}


def validate(akn_doc, strict=False):
    """ Validate this AKN document against its schema. if `strict` is True,
    then also validate the uniqueness of eId attributes. Returns a (validates, errors)
    tuple.
    """
    schema = get_schema(akn_doc.namespace, strict)
    return validate_xml(akn_doc.root, schema)


def validate_xml(root, schema):
    """ Validate an XML tree against a schema, returning a (validates, errors) tuple.
    """
    validates = schema(root)
    if not validates:
        errors = schema.error_log
    else:
        errors = []
    return validates, errors


def get_schema(namespace, strict):
    """ Get an XML schema for a namespace.
    """
    if not strict:
        namespace = namespace + '-lenient'

    if namespace not in _schemas:
        fname = os.path.join(os.path.dirname(__file__), 'xsd', SCHEMAS[namespace])
        with open(fname) as f:
            _schemas[namespace] = etree.XMLSchema(etree.parse(f))

    return _schemas[namespace]


def assert_validates(akn_doc, strict=False):
    """ Assert that this AKN document validates against the AKN schema.
    Raises `lxml.etree.DocumentInvalid` if validation fails.
    """
    schema = get_schema(akn_doc.namespace, strict)
    schema.assertValid(akn_doc.root)


class AkomaNtoso30:
    """ Information on various elements of the Akoma Ntoso 3.0 schema.
    """

    hier_elements = [
        'alinea', 'article', 'book', 'chapter', 'clause', 'division', 'indent', 'level', 'list', 'paragraph', 'part',
        'point', 'proviso', 'rule', 'section', 'subchapter', 'subclause', 'subdivision', 'sublist', 'subparagraph',
        'subpart', 'subrule', 'subsection', 'subtitle', 'title', 'tome', 'transitional'
    ]
    """ Hierarchical elements """



from .akn import StructuredDocument
from .collection import CollectionStructure


class AmendmentStructure(StructuredDocument):
    structure_type = "amendmentStructure"
    main_content_tag = "amendmentBody"


class Amendment(AmendmentStructure):
    document_type = "amendment"


class AmendmentList(CollectionStructure):
    document_type = "amendmentList"



from .akn import StructuredDocument


class DebateStructure(StructuredDocument):
    structure_type = "debateStructure"
    main_content_tag = "debateBody"


class Debate(DebateStructure):
    document_type = "debate"

    @classmethod
    def empty_document_content(cls, E):
        return E('debateBody',
                 E('debateSection',
                   E('p', eId="dbsect_nn_1__p_1")), eId="dbsect_nn_1")



from .akn import AkomaNtosoDocument, StructuredDocument, datestring
from .amendment import AmendmentStructure, Amendment, AmendmentList
from .collection import CollectionStructure, Collection, OfficialGazette
from .debate import DebateStructure, Debate
from .hierarchical import HierarchicalStructure, Act, AmendmentEvent, RepealEvent, Bill
from .judgment import JudgmentStructure, Judgment
from .openstructure import OpenStructure, DebateReport, Document, Statement
from .portion import PortionStructure, Portion
from .uri import FrbrUri

__all__ = [
    'Act', 'AkomaNtosoDocument', 'Amendment', 'AmendmentEvent', 'AmendmentList', 'AmendmentStructure',
    'Bill',
    'Collection', 'CollectionStructure',
    'Debate', 'DebateReport', 'DebateStructure', 'Document', 'datestring',
    'FrbrUri',
    'HierarchicalStructure',
    'Judgment', 'JudgmentStructure',
    'OfficialGazette', 'OpenStructure',
    'Portion', 'PortionStructure',
    'RepealEvent',
    'Statement', 'StructuredDocument',
]



import re

FRBR_URI_RE = re.compile(r"""^(/(?P<prefix>akn))?            # optional 'akn' prefix
                              /(?P<country>[a-z]{2})         # country
                              (-(?P<locality>[^/]+))?        # locality code
                              /(?P<doctype>[^/]+)            # document type
                              (/(?P<subtype>[^0-9][^/]*))?   # subtype (optional, cannot start with a number)
                              (/(?P<actor>[^0-9][^/]*))?     # actor (optional), cannot start with a number
                              /(?P<date>[0-9]{4}(-[0-9]{2}(-[0-9]{2})?)?)  # date
                              /(?P<number>[^/]+)             # number
                              (/                             # optional expression language and date
                                  (?P<language>[a-z]{3})                    # language (eg. eng)
                                  (?P<expression_date>[@:][^/]*)?           # expression date (eg. @ or @2012-12-22 or :2012-12-22)
                              )?
                              (/
                                  (!(?P<work_component>[^~.]+?))?           # optional component (eg. !main or !schedule1)
                                  (~(?P<portion>[^.]+))?                    # optional portion
                              )?
                              (\.(?P<format>[a-z0-9]+))?     # optional format (eg. .xml, .akn, .html, .pdf)
                              $""", re.X)


class FrbrUri(object):
    """
    An object for working with
    `Akoma Ntoso 3.0 FRBR URIs <https://docs.oasis-open.org/legaldocml/akn-nc/v1.0/os/akn-nc-v1.0-os.html>`_ (IRIs).

    URIs can be parsed from a plain string using :meth:`parse` or they can be
    constructed directly. URIs can be manipulated in-place once constructed,
    which is useful for turning a work-level URI into an expression or
    manifestation URI.

    URIs can be transformed back to the string form of work, expression and
    manifestation URIs using the :meth:`work_uri`, :meth:`expression_uri` and
    :meth:`manifestation_uri` methods.

    Example::

        >>> uri = FrbrUri.parse('/akn/za-jhb/act/by-law/2003/public-health/eng:2015-01-01/!main~part_1.xml')
        >>> uri.prefix
        'akn'
        >>> uri.country
        'za'
        >>> uri.locality
        'jhb'
        >>> uri.doctype
        'act'
        >>> uri.subtype
        'by-law'
        >>> uri.date
        '2003'
        >>> uri.number
        'public-health'
        >>> uri.language
        'eng'
        >>> uri.expression_date
        ':2015-01-01'
        >>> uri.work_component
        'main'
        >>> uri.portion
        'part_1'
        >>> uri.format
        'xml'
        >>> uri.work_uri()
        '/za-jhb/act/by-law/2003/public-health'
        >>> uri.expression_uri()
        '/za-jhb/act/by-law/2003/public-health/eng:2015-01-01/!main~part_1'
        >>> uri.manifestation_uri()
        '/za-jhb/act/by-law/2003/public-health/eng:2015-01-01/!main~part_1.xml'

    :ivar prefix: optional `akn` prefix
    :ivar country: two letter country code
    :ivar locality: locality within the country, may be None
    :ivar doctype: type of document (eg. ``act``)
    :ivar subtype: document subtype, may be None
    :ivar actor: emenating actor, may be None
    :ivar date: document data (str), YYYY[-MM[-DD]]
    :ivar number: document number (str)
    :ivar work_component: name of the work component, may be None
    :ivar language: three-letter expression language code, may be None
    :ivar expression_date: expression date (str), [@:]YYYY[-MM[-DD]], may be None
    :ivar format: format extension, may be None
    """

    default_language = 'eng'

    def __init__(self, country, locality, doctype, subtype, actor, date, number, work_component=None, language=None,
                 expression_date=None, format=None, portion=None, prefix="akn"):
        self.prefix = prefix
        self.country = country
        self.locality = locality
        self.doctype = doctype
        self.subtype = subtype
        self.actor = actor
        self.date = date
        self.number = number
        self.work_component = work_component
        self.portion = portion

        self.language = language or self.default_language
        self.expression_date = expression_date
        self.format = format

    def clone(self):
        """ Return a copy of this FrbrUri object.
        """
        return FrbrUri(
            prefix=self.prefix,
            country=self.country,
            locality=self.locality,
            doctype=self.doctype,
            subtype=self.subtype,
            actor=self.actor,
            date=self.date,
            number=self.number,
            work_component=self.work_component,
            language=self.language,
            expression_date=self.expression_date,
            portion=self.portion,
            format=self.format,
        )

    def uri(self):
        """ String form of the work URI, excluding the work component, if any. """
        return self.work_uri(work_component=False)

    def work_uri(self, work_component=True):
        """ String form of the work URI. """
        country = self.country
        parts = ['']
        if self.locality:
            country = country + "-" + self.locality

        if self.prefix:
            parts.append(self.prefix)

        parts += [country, self.doctype]

        if self.subtype:
            parts.append(self.subtype)
            if self.actor:
                parts.append(self.actor)

        parts += [self.date, self.number]

        if work_component and self.work_component:
            parts += ['!' + self.work_component]

        return '/'.join(parts)

    def expression_uri(self, work_component=True):
        """ String form of the expression URI. """
        if not self.language:
            raise ValueError("Expression URI requires a language.")

        uri = self.work_uri(work_component=False) + "/" + self.language

        if self.expression_date is not None:
            uri = uri + self.expression_date

        # if we have a work component, use it
        slashed = False
        if work_component and self.work_component:
            slashed = True
            uri = uri + "/!" + self.work_component

        if self.portion:
            if not slashed:
                uri = uri + "/"
            uri = uri + "~" + self.portion

        return uri

    def manifestation_uri(self, work_component=True):
        """ String form of the manifestation URI. """
        uri = self.expression_uri(work_component)
        if self.format:
            uri = uri + "." + self.format
        return uri

    def __str__(self):
        if self.format:
            return self.manifestation_uri()
        if self.expression_date or self.work_component:
            return self.expression_uri()
        return self.work_uri()

    def __repr__(self):
        return f'<FrbrUri({self})>'

    @classmethod
    def parse(cls, s):
        """ Parse a string into an FrbrUri instance.

        :raises ValueError: if parsing fails
        """
        s = s.rstrip('/')
        match = FRBR_URI_RE.match(s)
        if match:
            return cls(**match.groupdict())
        else:
            raise ValueError("Invalid FRBR URI: %s" % s)

    @property
    def year(self):
        """ The year, derived from :data:`date`. Read-only. """
        return self.date.split("-", 1)[0]

    @property
    def place(self):
        """ Full place code, including both country and locality (if present).
        """
        if self.locality:
            return self.country + "-" + self.locality
        return self.country



from .akn import StructuredDocument, datestring, NULL_DATE, parsedate


class HierarchicalStructure(StructuredDocument):
    structure_type = "hierarchicalStructure"
    main_content_tag = "body"


class Act(HierarchicalStructure):
    """
    An Akoma Ntoso Act document.
    """
    document_type = "act"

    @classmethod
    def empty_document_content(cls, E):
        return E('body',
                 E('section',
                   E('content',
                     E('p', eId="sec_nn_1__p_1")),
                   eId="sec_nn_1")
                 )

    @property
    def publication_name(self):
        """ Name of the publication in which this act was published.
        """
        pub = self.get_element('meta.publication')
        return pub.get('name') if pub is not None else None

    @publication_name.setter
    def publication_name(self, value):
        value = value or ""
        pub = self.ensure_publication()
        pub.set('name', value)
        pub.set('showAs', value)

    @property
    def publication_date(self):
        """ Date of publication as a `datetime.date` object.
        """
        pub = self.get_element('meta.publication')
        if pub is not None and pub.get('date'):
            return parsedate(pub.get('date'))
        return None

    @publication_date.setter
    def publication_date(self, value):
        self.ensure_publication().set('date', datestring(value))

    @property
    def publication_number(self):
        """ Sequence number of the publication.
        """
        pub = self.get_element('meta.publication')
        return pub.get('number') if pub is not None else None

    @publication_number.setter
    def publication_number(self, value):
        self.ensure_publication().set('number', value or "")

    @property
    def amendments(self):
        amendments = []

        for e in self.meta.iterfind(f'.//{{{self.namespace}}}lifecycle/{{{self.namespace}}}eventRef[@type="amendment"]'):
            date = parsedate(e.get('date'))
            event = AmendmentEvent(date=date)
            amendments.append(event)

            eid = e.get('source')[1:]
            source = self.meta.findall(f'.//{{{self.namespace}}}references/{{{self.namespace}}}passiveRef[@eId="{eid}"]')
            if source:
                event.amending_title = source[0].get('showAs')
                event.amending_uri = source[0].get('href')

        amendments.sort(key=lambda a: a.date)
        return amendments

    @amendments.setter
    def amendments(self, value):
        # delete existing entries
        lifecycle = self.meta.find(f'{{{self.namespace}}}lifecycle')
        if lifecycle is not None:
            for e in lifecycle.findall(f'./{{{self.namespace}}}eventRef[@type="amendment"]'):
                # delete the passive ref elements
                eid = e.get('source')[1:]
                for node in self.meta.iterfind(f'.//{{{self.namespace}}}references/{{{self.namespace}}}passiveRef[@eId="{eid}"]'):
                    node.getparent().remove(node)
                # delete the event
                lifecycle.remove(e)

        if not value:
            # no amendments, default is originalVersion so it doesn't need to be set explicitly
            if 'contains' in self.act.attrib:
                del self.act.attrib['contains']
            # lifecycle cannot be empty§
            if lifecycle is not None and not lifecycle.getchildren():
                lifecycle.getparent().remove(lifecycle)

        else:
            self.act.set('contains', 'singleVersion')
            lifecycle = self._ensure_lifecycle()
            references = self.ensure_element('meta.references', after=lifecycle)
            if not references.get('source'):
                references.set('source', '#' + self.source[1])

            for i, event in enumerate(value):
                date = datestring(event.date)
                ref = f'amendment-{i}-source'

                # create the lifecycle element
                node = self.make_element('eventRef')
                node.set('eId', 'amendment-' + date)
                node.set('date', date)
                node.set('type', 'amendment')
                node.set('source', '#' + ref)
                lifecycle.append(node)

                # create the passive ref
                node = self.make_element('passiveRef')
                node.set('eId', ref)
                node.set('href', event.amending_uri)
                node.set('showAs', event.amending_title)
                references.append(node)

    @property
    def repeal(self):
        e = self.meta.find(f'.//{{{self.namespace}}}lifecycle/{{{self.namespace}}}eventRef[@type="repeal"]')
        if e is not None:
            date = parsedate(e.get('date'))
            event = RepealEvent(date=date)

            id = e.get('source')[1:]
            source = self.meta.findall(f'.//{{{self.namespace}}}references/{{{self.namespace}}}passiveRef[@eId="{id}"]')
            if source:
                event.repealing_title = source[0].get('showAs')
                event.repealing_uri = source[0].get('href')
            return event

    @repeal.setter
    def repeal(self, value):
        # delete existing entries
        for e in self.meta.iterfind(f'.//{{{self.namespace}}}lifecycle/{{{self.namespace}}}eventRef[@type="repeal"]'):
            # delete the passive ref elements
            id = e.get('source')[1:]
            for node in self.meta.iterfind(f'.//{{{self.namespace}}}references/{{{self.namespace}}}passiveRef[@eId="{id}"]'):
                node.getparent().remove(node)

            # delete the event
            e.getparent().remove(e)

        if value:
            lifecycle = self._ensure_lifecycle()
            references = self.ensure_element('meta.references', after=lifecycle)

            date = datestring(value.date)
            ref = 'repeal-source'

            # create the lifecycle element
            node = self.make_element('eventRef')
            node.set('eId', 'repeal-' + date)
            node.set('date', date)
            node.set('type', 'repeal')
            node.set('source', '#' + ref)
            lifecycle.append(node)

            # create the passive ref
            node = self.make_element('passiveRef')
            node.set('eId', ref)
            node.set('href', value.repealing_uri)
            node.set('showAs', value.repealing_title)
            references.append(node)
        else:
            try:
                if len(list(self.meta.lifecycle.iterchildren())) == 0:
                    # remove empty lifecycle
                    self.meta.remove(self.meta.lifecycle)
            except AttributeError:
                pass

    def ensure_publication(self):
        return self.ensure_element('meta.publication', after=self.meta.identification,
                                   attribs={'showAs': '', 'name': '', 'date': NULL_DATE})


class AmendmentEvent(object):
    """ An event that amended a document.

    :ivar date: :class:`datetime.date` date of the event
    :ivar amending_title: String title of the amending document
    :ivar amending_uri: String form of the FRBR URI of the amending document
    """
    def __init__(self, date=None, amending_title=None, amending_uri=None):
        self.date = date
        self.amending_title = amending_title
        self.amending_uri = amending_uri


class RepealEvent(object):
    """ An event that repealed a document.

    :ivar date: :class:`datetime.date` date of the event
    :ivar repealing_title: String title of the repealing document
    :ivar repealing_uri: String form of the FRBR URI of the repealing document
    """
    def __init__(self, date=None, repealing_title=None, repealing_uri=None):
        self.date = date
        self.repealing_title = repealing_title
        self.repealing_uri = repealing_uri


class Bill(HierarchicalStructure):
    """
    An Akoma Ntoso Bill document.
    """
    document_type = "bill"



from .akn import StructuredDocument


class OpenStructure(StructuredDocument):
    structure_type = "openStructure"
    main_content_tag = "mainBody"


class DebateReport(OpenStructure):
    document_type = "debateReport"


class Document(OpenStructure):
    document_type = "doc"


class Statement(OpenStructure):
    document_type = "statement"
